#include "fw/lewan/servo.h"

#include "fw/lewan/tester.h"
#include "fw/servo/base.h"

using namespace schpin;

servo_ptr
schpin::make_lewan_servo( lewan_servo_id id, clk_time now, std::pmr::memory_resource* mem )
{
        static_assert( servo_bucket_size >= sizeof( lewan_servo ) );
        SCHPIN_ASSERT( mem != nullptr );
        void* target = mem->allocate( sizeof( lewan_servo ), alignof( lewan_servo ) );
        return { new ( target ) lewan_servo{ id, now }, servo_deleter{ mem } };
}
servo_ptr
schpin::make_lewan_test_servo( lewan_servo_id id, clk_time now, std::pmr::memory_resource* mem )
{
        static_assert( servo_bucket_size >= sizeof( lewan_servo_tester ) );
        SCHPIN_ASSERT( mem != nullptr );
        void* target = mem->allocate( sizeof( lewan_servo_tester ), alignof( lewan_servo_tester ) );
        return { new ( target ) lewan_servo_tester{ id, now }, servo_deleter{ mem } };
}

void lewan_servo::stop( servo_visitor& vis )
{
        // TODO: DO NOT DO THIS IF ACTIVATING?
        task_ = idle_task{};
        transmit_servo< LEWAN_SERVO_MOVE_STOP, reply::NO >( vis );
}

std::optional< servo_master_message >
lewan_servo::on_servo_receive( clk_time clock_time, s_ctl_id, const servo_receive_variant& svar )
{
        const auto& var = std::get< lewan_master_variant >( svar );
        lewan_catcher_.check( var );
        context_.on_servo_receive( clock_time, var );

        using pos_tuple = std::tuple< cmd_tag< LEWAN_SERVO_POS_READ >, lewan_full_angle >;
        if ( !std::holds_alternative< pos_tuple >( var ) ) {
                return {};
        }

        new_pos_ = true;

        lewan_angle change = config_map_.get_val< LEWAN_ANGLE_CHANGE_REPORT >();
        if ( change == lewan_angle::get< 0 >() ) {
                return {};
        }

        if ( angle_dist( last_report_, context_->pos ) < *change ) {
                return {};
        }
        last_report_ = context_->pos;
        return servo_master_group::create< POSITION >( converter_.se_to_in( context_->pos ) );
}

void lewan_servo::tick( clk_time clock_time, servo_visitor& vis )
{
        // TODO:
        //  STALL DETECTION
        if ( new_pos_ ) {
                new_pos_ = false;
                if ( pos_pid_enabled_ ) {
                        float actual_pose = static_cast< float >( *context_->pos );
                        pos_pid_.update( clock_time, actual_pose, desired_pos_ );
                        desired_velocity_ = pos_pid_.get_output();
                        /*
                        SCHPIN_INFO_LOG(
                            "servo",
                            "id: " << id_ << " actual: " << actual_pose << " desired: "
                                   << desired_pos_ << " out_vel: " << desired_velocity_ );
                                   */
                        servo_reply_ctl< TELEM >(
                            vis,
                            bangle::from_float( actual_pose ),
                            bangle::from_float( desired_pos_ ),
                            bangular_velocity::from_float( desired_velocity_ ) );
                }
                if ( vel_pid_enabled_ ) {
                        float vel = static_cast< float >( *context_->velocity );
                        vel_pid_.update( clock_time, vel, desired_velocity_ );
                        /*
                        SCHPIN_INFO_LOG(
                            "servo",
                            "id: " << id_ << " vel: " << vel << " desired: " << desired_velocity_
                                   << " out: " << vel_pid_.get_output() );
                                   */
                        transmit_servo< LEWAN_SERVO_OR_MOTOR_MODE_WRITE, reply::NO >(
                            vis,
                            lewan_motor_mode,
                            uint8_t{ 0 },
                            *lewan_angular_velocity::make( vel_pid_.get_output() ) );
                }
        }
        task_ << match(
            [&]( idle_task ) {},
            [&]( const activate_task& s ) {
                    activate_tick( s, vis );
            },
            [&]( goto_task t ) {
                    // stuff may overflow badyl here
                    float goal_f      = static_cast< float >( *t.goal );
                    float start_f     = static_cast< float >( *t.start );
                    float used_time_f = static_cast< float >( em::min(
                        *clock_time - *t.start_time, static_cast< uint32_t >( *t.time ) ) );
                    if ( used_time_f != 0.f ) {
                            desired_pos_ = em::map_range(
                                used_time_f,
                                0.f,
                                static_cast< float >( *t.time ),
                                start_f,
                                goal_f );
                    } else {
                            desired_pos_ = goal_f;
                    }
                    task_ = t;
                    if ( *t.start_time + static_cast< uint32_t >( *t.time ) > *clock_time ) {
                            return;
                    }
                    // random number - meh TODO
                    if ( angle_dist( context_->pos, t.goal ) < 15 ) {
                            task_ = idle_task{};
                    }
            },
            [&]( const spin_task& t ) {
                    if ( t.clock_end > clock_time ) {
                            return;
                    }
                    desired_velocity_ = 0.f;
                    task_             = idle_task{};
            } );
}

void lewan_servo::on_event(
    ctl_id                      cid,
    clk_time                    clock_time,
    const servo_events_variant& svar,
    servo_visitor&              vis )
{
        // TODO: error handling
        const auto& var = std::get< lewan_servo_events_variant >( svar );
        // TODO: the states are based on velocity controll in each instance -> to simplify
        // implementation/problems switch to velocity pid control in all of them
        //  - this simplifies setup, as it removes a lot of PID problems

        var << match(
            [&]( lewan_goto_event e ) {
                    vel_pid_enabled_ = true;
                    vel_pid_.set_time( clock_time );

                    pos_pid_enabled_ = true;
                    pos_pid_.set_time( clock_time );

                    desired_pos_ = *context_->pos;
                    task_        = goto_task{
                        context_->pos,
                        e.g,
                        clock_time,
                        e.t,
                    };
            },
            [&]( lewan_spin_event e ) {
                    desired_velocity_ = static_cast< float >( *e.av );

                    vel_pid_enabled_ = true;
                    vel_pid_.set_time( clock_time );

                    pos_pid_enabled_ = false;
                    pos_pid_.reset();

                    task_ = spin_task{ clock_time + clk_time{ e.t } };
            },
            [&]( lewan_config_event e ) {  //
                    auto opt_error = config_map_.set_val( e.addr, e.buffr );
                    if ( opt_error ) {
                            servo_reply_ctl_error< CONFIG_E >( cid, vis );
                    }
                    configure_pid();
                    converter_ = lewan_converter{ config_map_ };
            } );
}

void lewan_servo::poll( clk_time, std::size_t i_offset, servo_visitor& vis )
{
        poll_i_       = ( poll_i_ + 1 ) % 13;
        std::size_t i = ( i_offset + poll_i_ ) % 13;
        switch ( i ) {
                case 4:
                        transmit_servo< LEWAN_SERVO_VIN_READ, reply::YES >( vis );
                        break;
                case 8:
                        transmit_servo< LEWAN_SERVO_TEMP_READ, reply::YES >( vis );
                        break;
                default:
                        transmit_servo< LEWAN_SERVO_POS_READ, reply::YES >( vis );
                        break;
        }
}

em::either< servo_events_variant, servo_master_message >
lewan_servo::on_servo_packet( ctl_id id, const master_servo_command_variant& subcmd )
{
        using return_type = em::either< servo_events_variant, servo_master_message >;

        return subcmd << tuple_match(
                   [&]( cmd_tag< GOTO >, bangle a, btime t ) -> return_type {
                           return handle_in_to_se( converter_, id, GOTO, a )
                               .convert_left( [&]( lewan_angle an ) -> lewan_servo_events_variant {
                                       SCHPIN_INFO_LOG(
                                           "servo",
                                           "LEWAN " << id_ << " GOTO " << an << " (" << a << "), "
                                                    << t );
                                       return lewan_goto_event{ an, t };
                               } )
                               .construct_left< servo_events_variant >();
                   },
                   [&]( cmd_tag< SPIN >, bangular_velocity av, btime t ) -> return_type {
                           return handle_in_to_se( converter_, id, SPIN, av )
                               .convert_left(
                                   [&]( lewan_angular_velocity anv ) -> lewan_servo_events_variant {
                                           SCHPIN_INFO_LOG(
                                               "servo",
                                               "LEWAN " << id_ << " SPIN " << anv << " , " << t );
                                           return lewan_spin_event{ anv, t };
                                   } )
                               .construct_left< servo_events_variant >();
                   },
                   [&]( cmd_tag< POSITION > ) -> return_type {
                           return make_servo_message< POSITION >(
                               converter_.se_to_in( context_->pos ) );
                   },
                   [&]( cmd_tag< TEMP > ) -> return_type {
                           return make_servo_message< TEMP >(
                               converter_.se_to_in( context_->temperature ) );
                   },
                   [&]( cmd_tag< VOLTAGE > ) -> return_type {
                           return make_servo_message< VOLTAGE >(
                               converter_.se_to_in( context_->voltage ) );
                   },
                   [&]( cmd_tag< VELOCITY > ) -> return_type {
                           return make_servo_message< VELOCITY >(
                               converter_.se_to_in( context_->velocity ) );
                   },
                   [&]( cmd_tag< TASK > ) -> return_type {
                           servo_task_enum e = task_ << match(
                                                   [&]( idle_task ) {
                                                           return TASK_IDLE;
                                                   },
                                                   [&]( activate_task ) {
                                                           return TASK_ACTIVATE;
                                                   },
                                                   [&]( goto_task ) {
                                                           return TASK_GOTO;
                                                   },
                                                   [&]( spin_task ) {
                                                           return TASK_SPIN;
                                                   } );

                           return make_servo_message< TASK >( e );
                   },
                   [&]( cmd_tag< CONFIG_GET >, config_addr addr ) -> return_type {
                           return make_servo_message< CONFIG_GET >(
                               addr,
                               config_map_.get_buffered_val< config_buffer >(
                                   static_cast< lewan_config_enum >( addr ) ) );
                   },
                   [&]( cmd_tag< CONFIG_SET >,
                        config_addr   addr,
                        config_buffer buffr ) -> return_type {
                           return servo_events_variant{
                               lewan_servo_events_variant{ lewan_config_event{
                                   static_cast< lewan_config_enum >( addr ), buffr } } };
                   } );
}

void lewan_servo::dump_info()
{
#ifdef SCHPIN_USE_STREAMS
        std::cout << config_map_;
#endif
}
