// Copyright 2021 Schpin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include "magic_enum.hpp"
#include "schpin_lib/quantity.h"
#include "schpin_lib_exp/bounded.h"
#include "schpin_lib_exp/protocol/message.h"
#include "schpin_lib_exp/protocol/sizeless_message.h"
#include "schpin_lib_exp/scaled.h"

#include <bitset>
#include <emlabcpp/either.h>
#include <emlabcpp/iterators/numeric.h>
#include <variant>

#pragma once

namespace em = emlabcpp;

namespace schpin
{

struct argument_index : em::quantity< argument_index, int16_t >
{
        using em::quantity< argument_index, int16_t >::quantity;

        static std::string get_unit()
        {
                return "";
        }
};

enum protocol_error_enum : uint16_t
{
        INVALID_SIZE_E,
        MISSING_MATCH_E,
        BAD_ARGUMENT_E,
        RUNAWAY_TUPLE_E,
        UNKNOWN_REG_E
};

struct protocol_error_record
{
        protocol_error_enum type;
        uint16_t         byte_index;
};

inline std::ostream& operator<<( std::ostream& os, const protocol_error_record& rec )
{
        return os << "Error: " << magic_enum::enum_name( rec.type ) << " at: " << rec.byte_index;
}

template < auto ID >
struct cmd_tag
{
        using value_type                  = decltype( ID );
        static constexpr value_type value = ID;
};

template < auto ID >
inline std::ostream& operator<<( std::ostream& os, cmd_tag< ID > )
{
        return os << magic_enum::enum_name( cmd_tag< ID >::value );
}

struct value_tag
{
};
struct error_tag
{
};

template < typename T >
struct prot_res
{
        std::size_t                              used;
        std::variant< T, protocol_error_record > val_var;

        constexpr prot_res()
          : used( 0 )
          , val_var( T{} )
        {
        }

        prot_res( std::size_t u, T item )
          : used( u )
          , val_var( std::move( item ) )
        {
        }
        prot_res( std::size_t u, protocol_error_record e )
          : used( u )
          , val_var( e )
        {
        }
};

}  // namespace schpin

#ifdef SCHPIN_USE_NLOHMANN_JSON

template < auto ID >
struct nlohmann::adl_serializer< schpin::cmd_tag< ID > >
{
        static void to_json( nlohmann::json& j, schpin::cmd_tag< ID > )
        {
                j = magic_enum::enum_name( ID );
        }
        static void from_json( const nlohmann::json& j, schpin::cmd_tag< ID >& id )
        {

                auto opt_val = magic_enum::enum_cast< decltype( ID ) >( j.get< std::string >() );
                if ( !opt_val.has_value() ) {
                        throw nlohmann::json::parse_error::create( -1, 0, "Failed to parse ID" );
                }

                if ( opt_val.value() != ID ) {
                        throw nlohmann::json::parse_error::create( -1, 0, "ID does not match" );
                }

                id = schpin::cmd_tag< ID >{};
        }
};

#endif
