// Copyright 2021 Schpin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include "schpin_lib_exp/bounded.h"
#include "schpin_lib_exp/protocol/base.h"
#include "schpin_lib_exp/protocol/message.h"
#include "schpin_lib_exp/protocol/offset.h"
#include "schpin_lib_exp/protocol/sizeless_message.h"
#include "schpin_lib_exp/serialize.h"

#include <bitset>
#include <variant>

#pragma once

namespace schpin
{

template < typename T >
concept protocol_base_type = std::is_arithmetic_v< T > || std::is_enum_v< T >;

template < typename T >
struct protocol_item;

template < protocol_base_type T >
struct protocol_item< T >
{
        using input_type                  = T;
        using output_type                 = T;
        static constexpr std::size_t size = serializer< T >::size;

        static constexpr std::size_t
        serialize_at( static_view< size, uint8_t* > buffer, const T& item )
        {
                serializer< T >::template serialize< serialization_endianess< input_type > >(
                    buffer, item );
                return size;
        }

        static constexpr prot_res< output_type >
        deserialize( const em::view< const uint8_t* >& buffer )
        {
                auto opt_view = try_static_view< size >( buffer );
                if ( !opt_view ) {
                        return { size, protocol_error_record{ INVALID_SIZE_E, 0 } };
                }
                return {
                    size,
                    serializer< T >::template deserialize< serialization_endianess< output_type > >(
                        *opt_view ) };
        }
};

template < std::size_t N >
struct protocol_item< std::bitset< N > >
{
        using input_type                  = std::bitset< N >;
        using output_type                 = std::bitset< N >;
        static constexpr std::size_t size = ( N + 7 ) / 8;

        static constexpr std::size_t
        serialize_at( static_view< size, uint8_t* > buffer, const std::bitset< N >& item )
        {
                for ( std::size_t i : em::range( size ) ) {
                        std::bitset< 8 > byte;
                        for ( std::size_t j : em::range< std::size_t >( 8u ) ) {
                                byte[j] = item[i * 8 + j];
                        }
                        buffer[i] = static_cast< uint8_t >( byte.to_ulong() );
                }
                return size;
        }

        static constexpr prot_res< output_type >
        deserialize( const em::view< const uint8_t* >& buffer )
        {
                std::bitset< N > res;
                for ( std::size_t i : em::range( size ) ) {
                        std::bitset< 8 > byte = buffer[i];
                        for ( std::size_t j : em::range< std::size_t >( 8u ) ) {
                                res[i * 8 + j] = byte[j];
                        }
                }
                return { size, res };
        }
};

template < typename MsgCounterType, std::size_t N, typename Owner >
struct protocol_item< prot_message< MsgCounterType, N, Owner > >
{
        using input_type                  = prot_message< MsgCounterType, N, Owner >;
        using output_type                 = prot_message< MsgCounterType, N, Owner >;
        using sub_protocol                = protocol_item< std::array< uint8_t, N > >;
        using id_protocol                 = protocol_item< MsgCounterType >;
        static constexpr std::size_t size = sub_protocol::size + id_protocol::size;

        static_assert( N < std::numeric_limits< MsgCounterType >::max() );

        static constexpr std::size_t serialize_at(
            static_view< size, uint8_t* >                   buffer,
            const prot_message< MsgCounterType, N, Owner >& m )
        {
                std::size_t used =
                    id_protocol::serialize_at( buffer, static_cast< MsgCounterType >( m.size() ) );

                for ( std::size_t i : em::range( m.size() ) ) {
                        buffer[i + used] = m[i];
                }
                return m.size() + used;
        }
        static constexpr prot_res< output_type >
        deserialize( const em::view< const uint8_t* >& buffer )
        {
                if ( buffer.empty() ) {
                        return { 0, protocol_error_record{ INVALID_SIZE_E, 0 } };
                }

                auto [used, val_var] = id_protocol::deserialize( buffer );

                if ( std::holds_alternative< protocol_error_record >( val_var ) ) {
                        return { 0, std::get< protocol_error_record >( val_var ) };
                }
                MsgCounterType real_size = std::get< MsgCounterType >( val_var );

                if ( buffer.size() - used < real_size ) {
                        return { 0, protocol_error_record{ INVALID_SIZE_E, 0 } };
                }
                std::array< uint8_t, N > data{};
                for ( std::size_t i = 0; i < real_size; ++i ) {
                        data[i] = buffer[i + used];
                }

                return {
                    static_cast< std::size_t >( real_size ) + used,
                    output_type{ data, static_cast< std::size_t >( real_size ) } };
        }
};

// TODO: this needs a lot of testing

template < typename... Ts >
struct protocol_item< std::tuple< Ts... > >
{
        using item_tuple  = std::tuple< Ts... >;
        using input_type  = std::tuple< typename protocol_item< Ts >::input_type... >;
        using output_type = std::tuple< typename protocol_item< Ts >::output_type... >;

        static constexpr std::size_t size =
            em::sum( std::initializer_list< std::size_t >{ protocol_item< Ts >::size... } );

        static constexpr std::size_t
        serialize_at( static_view< size, uint8_t* > buffer, const input_type& tpl )
        {
                auto iter = buffer.begin();
                em::for_each( tpl, [&]( const auto& value ) {
                        using pitem   = protocol_item< std::decay_t< decltype( value ) > >;
                        auto opt_view = try_static_view< pitem::size >( iter, buffer.end() );
                        if ( opt_view ) {
                                iter += pitem::serialize_at( *opt_view, value );
                        } else {
                                // TODO: make error handling
                        }
                } );
                return static_cast< std::size_t >( iter - buffer.begin() );
        }

        static constexpr prot_res< output_type >
        deserialize( const em::view< const uint8_t* >& buffer )
        {
                std::tuple< std::optional< typename protocol_item< Ts >::output_type >... > tmp;

                auto                                   iter = buffer.begin();
                std::optional< protocol_error_record > opt_err;
                em::for_each( em::make_sequence_tuple_t< sizeof...( Ts ) >{}, [&]( auto index ) {
                        static constexpr std::size_t i = decltype( index )::value;
                        using pitem = protocol_item< std::tuple_element_t< i, item_tuple > >;

                        auto [used, val_var] = pitem::deserialize( em::view{ iter, buffer.end() } );

                        if ( std::holds_alternative< protocol_error_record >( val_var ) ) {
                                opt_err = std::get< protocol_error_record >( val_var );
                        } else {
                                std::get< i >( tmp ) = std::make_optional(
                                    std::get< typename pitem::output_type >( val_var ) );
                        }

                        std::advance( iter, used );
                } );

                std::size_t used = static_cast< std::size_t >( iter - buffer.begin() );

                if ( opt_err ) {
                        return { 0, *opt_err };
                }
                return std::apply(
                    [&]( auto... opt_items ) -> prot_res< output_type > {
                            return { used, output_type{ ( *opt_items )... } };
                    },
                    tmp );
        }
};

template < typename... Ts >
struct protocol_item< std::variant< Ts... > >
{
        using id_type = uint8_t;
        static_assert( sizeof...( Ts ) < std::numeric_limits< id_type >::max() );

        using input_type  = std::variant< Ts... >;
        using output_type = std::variant< Ts... >;

        static constexpr std::size_t size =
            sizeof( id_type ) + std::max( { protocol_item< Ts >::size... } );

        static constexpr std::size_t
        serialize_at( static_view< size, uint8_t* > buffer, const input_type& var )
        {
                buffer[0]     = static_cast< id_type >( var.index() );
                auto sub_buff = *try_static_view< size - 1 >( buffer.begin() + 1, buffer.end() );
                std::size_t sub_used =
                    var << match( [&]( const auto& item ) {
                            return protocol_item< std::decay_t< decltype( item ) > >::serialize_at(
                                sub_buff, item );
                    } );
                return sub_used + 1;
        }

        static constexpr prot_res< output_type >
        deserialize( const em::view< const uint8_t* >& buffer )
        {

                prot_res< output_type > res{ 0, protocol_error_record{ BAD_ARGUMENT_E, 0 } };

                em::for_each(
                    em::make_sequence_tuple_t< sizeof...( Ts ) >{}, [&]( auto integ_const ) {
                            using ic = decltype( integ_const );
                            using t  = std::variant_alternative_t< ic::value, output_type >;
                            if ( buffer[0] != ic::value ) {
                                    return;
                            }
                            res = protocol_item< t >::deserialize( em::tail( buffer ) );
                            res.used += 1;
                    } );
                return res;
        }
};

template < std::size_t N >
struct protocol_item< prot_sizeless_message< N > >
{
        using input_type                  = prot_sizeless_message< N >;
        using output_type                 = prot_sizeless_message< N >;
        static constexpr std::size_t size = N;

        static constexpr std::size_t
        serialize_at( static_view< size, uint8_t* > buffer, const prot_sizeless_message< N >& m )
        {
                for ( std::size_t i : em::range( m.size() ) ) {
                        buffer[i] = m[i];
                }
                return m.size();
        }

        static constexpr prot_res< output_type >
        deserialize( const em::view< const uint8_t* >& buffer )
        {
                if ( buffer.size() > N ) {
                        return { 0, protocol_error_record{ INVALID_SIZE_E, 0 } };
                }

                em::static_vector< uint8_t, N > res;
                for ( uint8_t val : buffer ) {
                        res.push_back( val );
                }

                return { res.size(), prot_sizeless_message< N >{ res } };
        }
};

template < typename T >
struct protocol_recursive_promoter
{
};

template < typename MsgCounterType, std::size_t N, typename Owner >
struct protocol_item< protocol_recursive_promoter< prot_message< MsgCounterType, N, Owner > > >
{
        using sub_protocol = protocol_item< prot_message< MsgCounterType, N, Owner > >;

        using input_type  = prot_message< MsgCounterType, N, Owner >;
        using output_type = typename Owner::output_type;

        static constexpr std::size_t size = sub_protocol::size;

        static constexpr std::size_t serialize_at(
            static_view< size, uint8_t* >                   buffer,
            const prot_message< MsgCounterType, N, Owner >& n )
        {
                return sub_protocol::serialize_at( buffer, n );
        }

        static constexpr prot_res< output_type >
        deserialize( const em::view< const uint8_t* >& buffer )
        {
                auto [used, val_var] = sub_protocol::deserialize( buffer );

                if ( std::holds_alternative< protocol_error_record >( val_var ) ) {
                        return { used, std::get< protocol_error_record >( val_var ) };
                }

                prot_res< output_type > res{};
                res.used = used;
                Owner::extract( std::get< prot_message< MsgCounterType, N, Owner > >( val_var ) )
                    .match(
                        [&]( output_type v ) {
                                res.val_var = v;
                        },
                        [&]( auto e ) {
                                res.val_var = e;
                        } );
                return res;
        }
};

template < typename Tag, typename T, T Offset >
struct protocol_item< protocol_offset< Tag, T, Offset > >
{
        using input_type  = protocol_offset< Tag, T, Offset >;
        using output_type = protocol_offset< Tag, T, Offset >;

        static constexpr std::size_t    size = serializer< T >::size;
        static constexpr endianess_enum endianess =
            serialization_endianess< protocol_offset< Tag, T, Offset > >;

        static constexpr std::size_t
        serialize_at( static_view< size, uint8_t* > buffer, const input_type& item )
        {
                serializer< T >::template serialize< endianess >( buffer, *item + Offset );
                return size;
        }

        static constexpr prot_res< output_type >
        deserialize( const em::view< const uint8_t* >& buffer )
        {
                auto opt_view = try_static_view< size >( buffer );
                if ( !opt_view ) {
                        return { 0, protocol_error_record{ INVALID_SIZE_E, 0 } };
                }
                auto val = serializer< T >::template deserialize< endianess >( *opt_view );
                return { size, output_type{ val - Offset } };
        }
};

template < typename T >
struct protocol_item_quantity
{
        using input_type  = T;
        using output_type = T;
        using inner_type  = typename T::value_type;
        static_assert( std::is_trivial_v< inner_type > );
        static constexpr std::size_t size = serializer< inner_type >::size;

        static constexpr std::size_t
        serialize_at( static_view< size, uint8_t* > buffer, const T& item )
        {
                serializer< inner_type >::template serialize<
                    serialization_endianess< input_type > >( buffer, *item );
                return size;
        }

        static constexpr prot_res< output_type >
        deserialize( const em::view< const uint8_t* >& buffer )
        {
                auto opt_view = try_static_view< size >( buffer );
                if ( !opt_view ) {
                        return { 0, protocol_error_record{ INVALID_SIZE_E, 0 } };
                }
                inner_type item = serializer< inner_type >::template deserialize<
                    serialization_endianess< output_type > >( *opt_view );
                return { size, T{ item } };
        }
};

template < typename Tag, typename T >
struct protocol_item< em::tagged_quantity< Tag, T > >
  : protocol_item_quantity< em::tagged_quantity< Tag, T > >
{
};

template < typename Tag, typename T, typename Scale >
struct protocol_item< scaled< Tag, T, Scale > > : protocol_item_quantity< scaled< Tag, T, Scale > >
{
};
template <>
struct protocol_item< argument_index > : protocol_item_quantity< argument_index >
{
};

template < typename T, T Min, T Max >
struct protocol_item< bounded< T, Min, Max > >
{
        using input_type  = bounded< T, Min, Max >;
        using output_type = bounded< T, Min, Max >;
        static_assert( std::is_trivial_v< T > );
        static constexpr std::size_t size = serializer< T >::size;

        static constexpr std::size_t
        serialize_at( static_view< size, uint8_t* > buffer, const bounded< T, Min, Max >& item )
        {
                serializer< T >::template serialize< serialization_endianess< input_type > >(
                    buffer, *item );
                return size;
        }

        static constexpr prot_res< output_type >
        deserialize( const em::view< const uint8_t* >& buffer )
        {
                auto opt_view = try_static_view< size >( buffer );
                if ( !opt_view ) {
                        return { 0, protocol_error_record{ INVALID_SIZE_E, 0 } };
                }
                T item =
                    serializer< T >::template deserialize< serialization_endianess< output_type > >(
                        *opt_view );
                auto opt_val = bounded< T, Min, Max >::make( item );
                if ( !opt_val ) {
                        return { 0, protocol_error_record{ BAD_ARGUMENT_E, 0 } };
                }
                return { size, *opt_val };
        }
};

template < typename T, std::size_t N >
struct protocol_item< std::array< T, N > >
{
        static_assert( N < std::numeric_limits< std::size_t >::max() );

        using input_type                  = std::array< T, N >;
        using output_type                 = std::array< T, N >;
        static constexpr std::size_t size = protocol_item< T >::size * N;

        static constexpr std::size_t
        serialize_at( static_view< size, uint8_t* > buffer, const std::array< T, N >& item )
        {
                auto buffers = buffer.template split< N >();
                for ( std::size_t i = 0; i < N; ++i ) {
                        protocol_item< T >::serialize_at( buffers[i], item[i] );
                }
                return size;
        }

        static constexpr prot_res< output_type >
        deserialize( const em::view< const uint8_t* >& buffer )
        {
                std::array< T, N > res;
                auto               beg = buffer.begin();
                for ( std::size_t i = 0; i < N; i++ ) {
                        em::view actual_buff{ beg, buffer.end() };
                        auto [used, val_var] = protocol_item< T >::deserialize( actual_buff );
                        beg += used;
                        if ( std::holds_alternative< protocol_error_record >( val_var ) ) {
                                protocol_error_record rec =
                                    std::get< protocol_error_record >( val_var );
                                rec.byte_index += static_cast< uint16_t >( beg - buffer.begin() );
                                return { 0, rec };
                        }
                        res[i] = std::get< T >( val_var );
                }

                return { size, res };
        }
};

}  // namespace schpin
